/****************************************************** 
 * File  : main.cc   
 * Desc. : extNADL Generator of 8-15-24-35 puzzles for 
 *         BDDSearch version 0.5
 *
 *         Positions and goal state
 *
 *       Y
 *
 *         ----------     
 *       2 |  |  |  |    
 *         | 1| 2| 3|    
 *         ----------   
 *       1 |  |  |  |    
 *         | 4| 5| 6|   
 *         ----------   
 *       0 |  |  |  |    
 *         | 7| 8|  |   
 *         ----------   
 *           0  1  2  X   
 *
 *
 *         This domain is generated for backward universal 
 *         planning, thus H is the heuristic distance to
 *         the initial state, NOT the goal state
 *         Up and Down actions may have a left and right
 *         effect as well. However this will never lead to 
 *         delta h of -2 (such that the sum of min manhattan
 *         distances still is admissible)
 *
 * Author: Rune M. Jensen
 * Date  : 10/23/02
 ******************************************************/

#include <vector>
#include <set>
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


using namespace std; 


void  printusage() {
  printf("USAGE: Npuzzle [-hglos]\n");
  printf("       -h print usage                                   \n");
  printf("       -n NUM                                           \n");
  printf("          size of N-puzzle                              \n");
  printf("            8                                           \n");
  printf("           15 (default)                                 \n");
  printf("           24                                           \n");
  printf("           35                                           \n");
  printf("       -g NUM                                           \n");
  printf("          different heuristics                          \n");
  printf("          1: Sum of Manhattan distances (default)       \n");
  printf("       -l NUM                                           \n");
  printf("          nuumber of random permutations of goal state  \n");
  printf("          to make initial state (default 10)            \n");
  printf("       -s NUM                                           \n");
  printf("          seed for random generator                     \n");
  printf("          (default 130271)                              \n");
  printf("       -o FILE                                          \n");
  printf("          name of outfile (it will be prefixed with the \n");
  printf("          solution length and postpixed with [n].nadl)  \n");
  printf("                                                        \n");
  printf("Creates a random NADL problem of the tile-puzzle with   \n");
  printf("the European goal state and the given solution bound.   \n");
  printf("H is for backward search an thus is the heuristic       \n");
  printf("distance to the initial state (NOT THE GOAL STATE)      \n");
} 


int myAbs(int v) {
  if (v < 0) 
    return -v;
  else
    return v;
}

//IN
// x,y    : board position 
// j      : tile number of tile occupying that position
// size   : size of puzzle (3,4,5)
//OUT
// the manhattan distance of the tile to it goal position
int manhattanDist(int x,int y,int xi,int yi) {
  
  return ( myAbs(x - xi) + myAbs(y - yi) );
} 


// compute delta h for various non-det outcomes
int dhUp(int x,int y,int t,int* initX,int* initY) {
  int hOld = manhattanDist(x,y+1,initX[t],initY[t]);
  int hNew = manhattanDist(x,y,initX[t],initY[t]);
  return hNew - hOld;
}

int dhDown(int x,int y,int t,int* initX,int* initY) {
  int hOld = manhattanDist(x,y-1,initX[t],initY[t]);
  int hNew = manhattanDist(x,y,initX[t],initY[t]);
  return hNew - hOld;
}


int dhRight(int x,int y,int t,int* initX,int* initY) {
  int hOld = manhattanDist(x+1,y,initX[t],initY[t]);
  int hNew = manhattanDist(x,y,initX[t],initY[t]);
  return hNew - hOld;
}


int dhLeft(int x,int y,int t,int* initX,int* initY) {
  int hOld = manhattanDist(x-1,y,initX[t],initY[t]);
  int hNew = manhattanDist(x,y,initX[t],initY[t]);
  return hNew - hOld;
}
  


int dhUpRight(int x,int y,int t,int tR,int* initX,int* initY) {
  int hOld = manhattanDist(x,y+1,initX[t],initY[t]) + manhattanDist(x+1,y+1,initX[tR],initY[tR]);
  int hNew = manhattanDist(x,y,initX[t],initY[t]) + manhattanDist(x,y+1,initX[tR],initY[tR]);
  return hNew - hOld;
}



int dhUpLeft(int x,int y,int t,int tL,int* initX,int* initY) {
  int hOld = manhattanDist(x,y+1,initX[t],initY[t]) + manhattanDist(x-1,y+1,initX[tL],initY[tL]);
  int hNew = manhattanDist(x,y,initX[t],initY[t]) + manhattanDist(x,y+1,initX[tL],initY[tL]);
  return hNew - hOld;
}

int dhDownRight(int x,int y,int t,int tR,int* initX,int* initY) {
  int hOld = manhattanDist(x,y-1,initX[t],initY[t]) + manhattanDist(x+1,y-1,initX[tR],initY[tR]);
  int hNew = manhattanDist(x,y,initX[t],initY[t]) + manhattanDist(x,y-1,initX[tR],initY[tR]);
  return hNew - hOld;
}

int dhDownLeft(int x,int y,int t,int tL,int* initX,int* initY) {
  int hOld = manhattanDist(x,y-1,initX[t],initY[t]) + manhattanDist(x-1,y-1,initX[tL],initY[tL]);
  int hNew = manhattanDist(x,y,initX[t],initY[t]) + manhattanDist(x,y-1,initX[tL],initY[tL]);
  return hNew - hOld;
}



///////////////////////////////////////////////////
//
// Main function
//
///////////////////////////////////////////////////
int main(int argc, char **argv) {
  
  // heuristic type
  int n = 15;
  int size = 4;
  int heuristic = 1;
  int solutionLengthBound = 10;
  int seed = 130271;

  char outName[512] = "Puzzle";
  char finalOutName[512];


  /*** Scan command line ***/
  for (int ind = 1; ind < argc; ind++) {
    
    /*** Parse switches ***/
    if (argv[ind][0] == '-') {
      switch (argv[ind][1]) {  
      case 'n': 
	n = atoi(argv[++ind]);
	size = int(sqrt(n+1));
	break;
      case 'g': 
	heuristic = atoi(argv[++ind]);
	break;
      case 'l': 
	solutionLengthBound = atoi(argv[++ind]);
	break;
      case 's': 
	seed = atoi(argv[++ind]);
	break;
      case 'o': 
	strcpy(outName,argv[++ind]);
	break;
      case 'h': 
	printusage();
	exit(1);
	break;
      default : printf("main.cc main : Unknown switch '%c'\nexiting\n", argv[ind][1]);
	exit(1);
	break;
      }
    }
  }


  // 1) Compute the initial state by a randomly moving the empty space 
  //    solutionLengthBound times
  
  // enter goal state
  int* goalX = new int[n+2];
  int* goalY = new int[n+2];

  // pos encodes init state
  int **pos = new int*[size];  
  for (int x = 0; x < size; x++)
    pos[x] =  new int[size];

  int tile = n + 1;
  for (int y = 0; y < size; y++)
    for (int x = size-1; x >= 0; x--)
      {
	pos[x][y] = tile;
	goalX[tile] = x;
	goalY[tile] = y;
	tile--;
      }

  pos[size - 1][0] = 0;
  goalX[0] = size - 1;
  goalY[0] = 0;

  int zeroTileX = size - 1;
  int zeroTileY = 0;
  int lastMove = 1;
  set<int> moves;

  srand(seed);

  // randomly move empty tile around
  for (int i = 0; i < solutionLengthBound; i++)
    {
      // 0 : up
      // 1 : down
      // 2 : right
      // 3 : left
      moves.clear();
      
      if (zeroTileY < size - 1)
	moves.insert(0);
      if (zeroTileY > 0)
	moves.insert(1);
      if (zeroTileX < size - 1)
	moves.insert(2);
      if (zeroTileX > 0)
	moves.insert(3);

      moves.erase(lastMove);
      
      // pick a random move
      int shuffle = rand() % moves.size();
      set<int>::iterator si = moves.begin();
      for (int s = 0; s < shuffle; s++)
	++si;
      
      int move = *si;

      lastMove = move;

      // perform state change
      switch (move) 
	{
	case 0: // up
	  pos[zeroTileX][zeroTileY] = pos[zeroTileX][zeroTileY+1];
	  pos[zeroTileX][zeroTileY+1] = 0; 
	  zeroTileY++;
	  break;
	case 1: // down
	  pos[zeroTileX][zeroTileY] = pos[zeroTileX][zeroTileY-1];
	  pos[zeroTileX][zeroTileY-1] = 0; 
	  zeroTileY--;
	  break;
	case 2: // right
	  pos[zeroTileX][zeroTileY] = pos[zeroTileX+1][zeroTileY];
	  pos[zeroTileX+1][zeroTileY] = 0; 
	  zeroTileX++;
	  break;
	case 3: // down
	  pos[zeroTileX][zeroTileY] = pos[zeroTileX-1][zeroTileY];
	  pos[zeroTileX-1][zeroTileY] = 0; 
	  zeroTileX--;
	  break;
	}
    }
  
  int* initX = new int[n + 1];
  int* initY = new int[n + 1];
  for (int x = 0; x < size; x++)
    for (int y = 0; y < size; y++)
      {
	initX[pos[x][y]] = x;
	initY[pos[x][y]] = y;
      }
    

  // find heuristic value of init
  // sum of Manhattan distances to the goal 
  int Hgoal = 0;
  for (int i = 1; i < n + 1; i++)
    {
      Hgoal += myAbs(goalX[i] - initX[i]);
      Hgoal += myAbs(goalY[i] - initY[i]);
    }
  
  sprintf(finalOutName,"%d%s%d.nadl",solutionLengthBound,outName,n);      
  ofstream  out(finalOutName,ios::out);
  
  out << "\% File:   " << finalOutName << endl;
  out << "\% Desc:   " << n << "-Puzzle problem with solution length upper bound of " 
      <<  solutionLengthBound << endl;
  out << "\%         Initial state:                                 \n"; 
  out << "\%     \n"; 
  for (int y = size - 1; y >= 0; y--)
    {
      out << "\%         ";
      for (int x = 0; x < size; x++)
	out << pos[x][y] << " ";
      out << endl;
    }
  out << "\%         \n"; 
  out << "\%         This encoding tracks the tile of each position\n"; 
  out << "\%         It's a nondet. version where up and down moves\n"; 
  out << "\%         of empty space may lead to a left and right move\n"; 
  out << "\%         as well. (however never such that h becomes -2)\n"; 
  out << "\%         H is for backward search an thus the distance \n"; 
  out << "\%         to the initial state (not the goal state)     \n"; 
  out << "\% Date:   02\n";
  out << "\% Auth:   Rune M. Jensen CS, CMU\n";
  
  out << "\nVARIABLES\n";
  
  out << "  nat(" << int(ceil(log(size*size)/log(2))) << ") ";
    for (int x = 0; x < size; x++)
      for (int y = 0; y < size; y++)
	if (x == 0 && y == 0)
	  out << "P" << x << y;
	else
	  out << ",P" << x << y;  
      
  
  
  out << "\n\nSYSTEM\n\n";
  
  // make all up moves
  out << " agt:  sys\n\n";  

  out << "  Up\n";
  // go through each position
  for (int x = 0; x < size; x++)
    for (int y = 0; y < (size - 1); y++)
      // go through each tile in turn
      for (int t = 1; t < (n + 1); t++)
	{
	  if (-dhUp(x,y,t,initX,initY) > -2)
	    {
	      // transitions with no horizontal effect
	      out << "     dg: 1\n";
	      out << "     dh: " << -dhUp(x,y,t,initX,initY) << "\n";
	      out << "    mod: P" << x << y << ",P" << x << y+1 << endl;
	      out << "    pre: P" << x << y << " = 0 /\\ P" << x << y+1 << " = " << t << endl; 
	      out << "    eff: P" << x << y << "' = " << t <<" /\\ P" << x << y+1 << "' = 0\n\n";
	    }
	}


  for (int x = 0; x < size; x++)
    for (int y = 0; y < (size - 1); y++)
      // go through each tile in turn
      for (int t = 1; t < (n + 1); t++)
	{
	  // transitions with right horizontal effect
	  if (x < 2)
	    // go through all right tiles
	    for (int tR = 1; tR < (n + 1); tR++)
	      if (tR != t)
		if (-dhUpRight(x,y,t,tR,initX,initY) > -2)
		  {
		    out << "     dg: 1\n";		  
		    out << "     dh: " << -dhUpRight(x,y,t,tR,initX,initY) << "\n";
		    out << "    mod: P" << x << y << ",P" << x << y+1 << ",P" << x+1 << y+1 << endl;
		    out << "    pre: P" << x << y << " = 0 /\\ P" << x << y+1 << " = " << t 
			<< " /\\ P" << x+1 << y+1 << " = " << tR << endl; 
		    out << "    eff: P" << x << y << "' = " << t << " /\\ P" << x << y+1 << "' = " << tR
			<< " /\\ P" << x+1 << y+1 << "' = 0\n\n";
		  }
	}
  
  for (int x = 0; x < size; x++)
    for (int y = 0; y < (size - 1); y++)
      // go through each tile in turn
      for (int t = 1; t < (n + 1); t++)
	{	     	  
	  // transitions with left horizontal effect
	  if (x > 0)
	    // go through all left tiles
	    for (int tL = 1; tL < (n + 1); tL++)
	      if (tL != t)
		if (-dhUpLeft(x,y,t,tL,initX,initY) > -2)
		{
		  out << "     dg: 1\n";
		  out << "     dh: " << -dhUpLeft(x,y,t,tL,initX,initY) << "\n";
		  out << "    mod: P" << x << y << ",P" << x << y+1 << ",P" << x-1 << y+1 << endl;
		  out << "    pre: P" << x << y << " = 0 /\\ P" << x << y+1 << " = " << t 
		      << " /\\ P" << x-1 << y+1 << " = " << tL << endl; 
		  out << "    eff: P" << x << y << "' = " << t << " /\\ P" << x << y+1 << "' = " << tL
		      << " /\\ P" << x-1 << y+1 << "' = 0\n\n";
		}
	}
  
  
  
  
  
  // make all down moves
  out << "  Down\n";    
  for (int x = 0; x < size; x++)
    for (int y = 1; y < size; y++)
      // go through each tile in turn
      for (int t = 1; t < n+1; t++)
	if (-dhDown(x,y,t,initX,initY) > -2)
	{
	  // transitions with no horizontal effect
	  out << "     dg: 1\n";
	  out << "     dh: " << -dhDown(x,y,t,initX,initY) << "\n";
	  out << "    mod: P" << x << y << ",P" << x << y-1 << endl;
	  out << "    pre: P" << x << y << " = 0 /\\ P" << x << y-1 << " = " << t << endl; 
	  out << "    eff: P" << x << y << "' = " << t << " /\\ P" << x << y-1 << "' = 0\n\n";
	}  
	  
  for (int x = 0; x < size; x++)
    for (int y = 1; y < size; y++)
      // go through each tile in turn
      for (int t = 1; t < n+1; t++)
	{
	  // transitions with right horizontal effect
	  if (x < 2)
	    // go through all right tiles
	    for (int tR = 1; tR < (n + 1); tR++)
	      if (tR != t)
		if (-dhDownRight(x,y,t,tR,initX,initY) > -2)
		  { 
		    out << "     dg: 1\n";
		    out << "     dh: " << -dhDownRight(x,y,t,tR,initX,initY) << "\n";
		    out << "    mod: P" << x << y << ",P" << x << y-1 << ",P" << x+1 << y-1 << endl;
		    out << "    pre: P" << x << y << " = 0 /\\ P" << x << y-1 << " = " << t 
			<< " /\\ P" << x+1 << y-1 << " = " << tR <<  endl; 
		    out << "    eff: P" << x << y << "' = " << t << " /\\ P" << x << y-1 << "' = " << tR
			<< " /\\ P" << x+1 << y-1 << "' = 0\n\n";
		  }
	}	  
  
  for (int x = 0; x < size; x++)
    for (int y = 1; y < size; y++)
      // go through each tile in turn
      for (int t = 1; t < n+1; t++)
	{
	  // transitions with left horizontal effect
	  if (x > 0)
	    // go through all left tiles
	    for (int tL = 1; tL < (n + 1); tL++)
	      if (tL != t)
		if (-dhDownLeft(x,y,t,tL,initX,initY) > -2)
		  {
		    out << "     dg: 1\n";
		    out << "     dh: " << -dhDownLeft(x,y,t,tL,initX,initY) << "\n";
		    out << "    mod: P" << x << y << ",P" << x << y-1 << ",P" << x-1 << y-1 << endl;
		    out << "    pre: P" << x << y << " = 0 /\\ P" << x << y-1 << " = " << t 
			<< " /\\ P" << x-1 << y-1 << " = " << tL <<  endl; 
		    out << "    eff: P" << x << y << "' = " << t << " /\\ P" << x << y-1 << "' = " << tL
			<< " /\\ P" << x-1 << y-1 << "' = 0\n\n";		  
		  }	  
	}
  
  

    
  // make all right moves
  out << "  Right\n";  
  for (int x = 0; x < size - 1; x++)
    for (int y = 0; y < size; y++)
      // go through each tile in turn
      for (int t = 1; t < n + 1; t++)
	{
	  out << "     dg: 1\n";
	  out << "     dh: " << -dhRight(x,y,t,initX,initY) << "\n";
	  out << "    mod: P" << x << y << ",P" << x+1 << y << endl;
	  out << "    pre: P" << x << y << " = 0 /\\ P" << x+1 << y << " = " << t << endl; 
          out << "    eff: P" << x << y << "' = "<< t << " /\\ P" << x+1 << y << "' = 0\n\n";
	}
  
  

  // make all left moves
  out << "  Left\n";  
  for (int x = 1; x < size; x++)
    for (int y = 0; y < size; y++)
      // go through each tile in turn
      for (int t = 1; t < n + 1; t++)
	{
	  out << "     dg: 1\n";	    
	  out << "     dh: " << -dhLeft(x,y,t,initX,initY) << "\n";
	  out << "    mod: P" << x << y << ",P" << x-1 << y << endl;
	  out << "    pre: P" << x << y << " = 0 /\\ P" << x-1 << y << " = " << t << endl; 
          out << "    eff: P" << x << y << "' = "<< t << " /\\ P" << x-1 << y << "' = 0\n\n";
	}
  

 
	  
  out << "\n\nINITIALLY\n";
  // give positions init tiling
  for (int i = 0; i < n + 1; i++)
    if (i == 0)
      out << " P" << initX[i]  << initY[i] << " = " << i << endl;
    else
      out << " /\\ P" << initX[i]  << initY[i] << " = " << i << endl;
  
  
  out << "\n\nGOAL\n";
  // give positions goal tiling
  for (int i = 0; i < n + 1; i++)
    if (i == 0)
      out << " P" << goalX[i]  << goalY[i] << " = " << i << endl;
    else
      out << " /\\ P" << goalX[i]  << goalY[i] << " = " << i << endl;
  out << " heu: " << Hgoal << endl;
  
  return 0;
}
