/****************************************************** 
 * File  : main.cc   
 * Desc. : Generator of 15puzzles for BDDSearch 0.3
 *         in NADL 
 *
 *         Positions and goal state
 *
 *       Y
 *         -------------
 *       3 |  |  |  |  | 
 *         | 1| 2| 3| 4|
 *         -------------
 *       2 |  |  |  |  | 
 *         | 5| 6| 7| 8|
 *         -------------
 *       1 |  |  |  |  | 
 *         | 9|10|11|12|
 *         -------------
 *       0 |  |  |  |  | 
 *         |13|14|15|  |
 *         -------------
 *           0  1  2  3  X
 *
 *
 * Author: Rune M. Jensen
 * Date  : 5/23/01
 ******************************************************/

#include <vector>
#include <set>
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


using namespace std; 


void  printusage() {
  printf("USAGE: Npuzzle [-hglos]\n");
  printf("       -h print usage                                   \n");
  printf("       -n NUM                                           \n");
  printf("          size of N-puzzle                              \n");
  printf("            8                                           \n");
  printf("           15 (default)                                 \n");
  printf("           24                                           \n");
  printf("           35                                           \n");
  printf("       -g NUM                                           \n");
  printf("          different heuristics                          \n");
  printf("          1: Sum of Manhattan distances                 \n");
  printf("       -l NUM                                           \n");
  printf("          nuumber of random permutations of goal state  \n");
  printf("          to make initial state (default 10)            \n");
  printf("       -s NUM                                           \n");
  printf("          seed for random generator                     \n");
  printf("          (default 130271)                              \n");
  printf("       -o FILE                                          \n");
  printf("          name of outfile (it will be prefixed with the \n");
  printf("          solution length and postpixed with [n].nadl   \n");
  printf("          and .init (for single-state A*)               \n");
  printf("                                                        \n");
  printf("Creates a random NADL problem of the 15/24puzzle with   \n");
  printf("the European goal state and the given solution bound.   \n");
  printf("The heuristic value of the initial state is written     \n");
  printf("into the value of an environment action (the first)     \n");
} 


int MyAbs(int v) {
  if (v < 0) 
    return -v;
  else
    return v;
}
  


int main(int argc, char **argv) {
  
  // heuristic type
  int n = 15;
  int size = 4;
  int heuristic = 1;
  int solutionLengthBound = 10;
  int seed = 130271;

  char outName[512] = "Puzzle";
  char finalOutName[512];


  /*** Scan command line ***/
  for (int ind = 1; ind < argc; ind++) {
    
    /*** Parse switches ***/
    if (argv[ind][0] == '-') {
      switch (argv[ind][1]) {  
      case 'n': 
	n = atoi(argv[++ind]);
	size = int(sqrt(n+1));
	break;
      case 'g': 
	heuristic = atoi(argv[++ind]);
	break;
      case 'l': 
	solutionLengthBound = atoi(argv[++ind]);
	break;
      case 's': 
	seed = atoi(argv[++ind]);
	break;
      case 'o': 
	strcpy(outName,argv[++ind]);
	break;
      case 'h': 
	printusage();
	exit(1);
	break;
      default : printf("main.cc main : Unknown switch '%c'\nexiting\n", argv[ind][1]);
	exit(1);
	break;
      }
    }
  }


  // 1) Compute the initial state by a randomly moving the empty space 
  //    solutionLengthBound times
  
  // enter goal state
  int goalX[n+2];
  int goalY[n+2];

  // pos encodes init state
  int **pos = new int*[size];  
  for (int x = 0; x < size; x++)
    pos[x] =  new int[size];

  int tile = n + 1;
  for (int y = 0; y < size; y++)
    for (int x = size-1; x >= 0; x--)
      {
	pos[x][y] = tile;
	goalX[tile] = x;
	goalY[tile] = y;
	tile--;
      }

  pos[size - 1][0] = 0;
  goalX[0] = size - 1;
  goalY[0] = 0;

  int zeroTileX = size - 1;
  int zeroTileY = 0;
  int lastMove = 1;
  set<int> moves;

  srand(seed);

  // randomly move empty tile around
  for (int i = 0; i < solutionLengthBound; i++)
    {
      // 0 : up
      // 1 : down
      // 2 : right
      // 3 : left
      moves.clear();
      
      if (zeroTileY < size - 1)
	moves.insert(0);
      if (zeroTileY > 0)
	moves.insert(1);
      if (zeroTileX < size - 1)
	moves.insert(2);
      if (zeroTileX > 0)
	moves.insert(3);

      moves.erase(lastMove);
      
      // pick a random move
      int shuffle = rand() % moves.size();
      set<int>::iterator si = moves.begin();
      for (int s = 0; s < shuffle; s++)
	++si;
      
      int move = *si;

      lastMove = move;

      // perform state change
      switch (move) 
	{
	case 0: // up
	  pos[zeroTileX][zeroTileY] = pos[zeroTileX][zeroTileY+1];
	  pos[zeroTileX][zeroTileY+1] = 0; 
	  zeroTileY++;
	  break;
	case 1: // down
	  pos[zeroTileX][zeroTileY] = pos[zeroTileX][zeroTileY-1];
	  pos[zeroTileX][zeroTileY-1] = 0; 
	  zeroTileY--;
	  break;
	case 2: // right
	  pos[zeroTileX][zeroTileY] = pos[zeroTileX+1][zeroTileY];
	  pos[zeroTileX+1][zeroTileY] = 0; 
	  zeroTileX++;
	  break;
	case 3: // down
	  pos[zeroTileX][zeroTileY] = pos[zeroTileX-1][zeroTileY];
	  pos[zeroTileX-1][zeroTileY] = 0; 
	  zeroTileX--;
	  break;
	}
    }
  
  int initX[n + 1];
  int initY[n + 1];
  for (int x = 0; x < size; x++)
    for (int y = 0; y < size; y++)
      {
	initX[pos[x][y]] = x;
	initY[pos[x][y]] = y;
      }
    

  // find heuristic value of init
  // sum of Manhattan distances to the goal 
  int Hinit = 0;
  for (int i = 1; i < n + 1; i++)
    {
      Hinit += MyAbs(goalX[i] - initX[i]);
      Hinit += MyAbs(goalY[i] - initY[i]);
    }
  
  sprintf(finalOutName,"%d%s%d.nadl",solutionLengthBound,outName,n);      
  ofstream  out(finalOutName,ios::out);
  
  out << "\% File:   " << finalOutName << endl;
  out << "\% Desc:   " << n << "-Puzzle problem with solution length upper bound of " 
      <<  solutionLengthBound << endl;
  out << "\%         This encoding tracks the tile of each position\n"; 
  out << "\% Date:   02\n";
  out << "\% Auth:   Rune M. Jensen CS, CMU\n";
  
  out << "\nVARIABLES\n";
  
  out << "  nat(" << int(ceil(log(size*size)/log(2))) << ") ";
    for (int x = 0; x < size; x++)
      for (int y = 0; y < size; y++)
	if (x == 0 && y == 0)
	  out << "P" << x << y;
	else
	  out << ",P" << x << y;  
      
  
  
  out << "\n\nSYSTEM\n\n";
  
  out << " agt: sys\n\n";
  
  // make all up moves
  for (int x = 0; x < size; x++)
    for (int y = 0; y < (size - 1); y++)
      // go through each tile in turn
      for (int t = 1; t < (n + 1); t++)
	{
	  out << "  UpX" << x << "Y" << y << "Tile" << t << endl;
	  // write dh + 10 as the value of the action
	  if (y >= goalY[t])
	    out << "    val: 9\n";
	  else 
	    out << "    val: 11\n";	    
	  out << "    mod: P" << x << y << ",P" << x << y+1 << endl;
	  out << "    pre: P" << x << y << " = 0 /\\ P" << x << y+1 << " = " << t << endl; 
	  out << "    eff: P" << x << y << "' = " << t <<" /\\ P" << x << y+1 << "' = 0\n\n";
	}
  
  
  

  
  // make all down moves
  for (int x = 0; x < size; x++)
    for (int y = 1; y < size; y++)
      // go through each tile in turn
      for (int t = 1; t < n+1; t++)
	{
	  out << "  DownX" << x << "Y" << y << "Tile" << t << endl;
	  // write dh + 10 as the value of the action
	  if (y <= goalY[t])
	    out << "    val: 9\n";
	  else 
	    out << "    val: 11\n";	    
	  out << "    mod: P" << x << y << ",P" << x << y-1 << endl;
	  out << "    pre: P" << x << y << " = 0 /\\ P" << x << y-1 << " = " << t << endl; 
	  out << "    eff: P" << x << y << "' = " << t << " /\\ P" << x << y-1 << "' = 0\n\n";
	}
  
  

  
  
  
  // make all right moves
  for (int x = 0; x < size - 1; x++)
    for (int y = 0; y < size; y++)
      // go through each tile in turn
      for (int t = 1; t < n + 1; t++)
	{
	  out << "  RightX" << x << "Y" << y << "Tile" << t << endl;
	  // write dh + 10 as the value of the action
	  if (x >= goalX[t])
	    out << "    val: 9\n";
	  else 
	    out << "    val: 11\n";	    
	  out << "    mod: P" << x << y << ",P" << x+1 << y << endl;
	  out << "    pre: P" << x << y << " = 0 /\\ P" << x+1 << y << " = " << t << endl; 
          out << "    eff: P" << x << y << "' = "<< t << " /\\ P" << x+1 << y << "' = 0\n\n";
	}
  
  

  // make all left moves
  for (int x = 1; x < size; x++)
    for (int y = 0; y < size; y++)
      // go through each tile in turn
      for (int t = 1; t < n + 1; t++)
	{
	  out << "  LeftX" << x << "Y" << y << "Tile" << t << endl;
	  // write dh + 10 as the value of the action
	  if (x <= goalX[t])
	    out << "    val: 9\n";
	  else 
	    out << "    val: 11\n";	    
	  out << "    mod: P" << x << y << ",P" << x-1 << y << endl;
	  out << "    pre: P" << x << y << " = 0 /\\ P" << x-1 << y << " = " << t << endl; 
          out << "    eff: P" << x << y << "' = "<< t << " /\\ P" << x-1 << y << "' = 0\n\n";
	}
  

  
  out << "\n\nENVIRONMENT\n";
  out << " agt: env\n\n";
  out << "  Dummy\n";
  out << "    val: " << Hinit << endl;
  out << "    mod:\n";
  out << "    pre: true\n";
  out << "    eff: true\n";

	  
  out << "\n\nINITIALLY\n";
  // give positions init tiling
  for (int i = 0; i < n + 1; i++)
    if (i == 0)
      out << " P" << initX[i]  << initY[i] << " = " << i << endl;
    else
      out << " /\\ P" << initX[i]  << initY[i] << " = " << i << endl;
  
  
  out << "\n\nGOAL\n";
  // give positions goal tiling
  for (int i = 0; i < n + 1; i++)
    if (i == 0)
      out << " P" << goalX[i]  << goalY[i] << " = " << i << endl;
    else
      out << " /\\ P" << goalX[i]  << goalY[i] << " = " << i << endl;

  // write the .init file for the single Astar algorithm

  char initName[512];
  sprintf(initName,"%d%s%d.init",solutionLengthBound,outName,n);      
  ofstream  out2(initName,ios::out);

  // first enter the number of data
  out2 << n + 1 << endl;
  
  // write init pos in x-y order
  for (int x = 0; x < size; x++)
    for (int y = 0; y < size; y++)
      out2 << pos[x][y] << endl;
  
  return 0;
}
